/**
* Copyright (c) 2008 Moritz Bechler
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
* copies of the Software, and to permit persons to whom the Software is
* furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in
* all copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
* THE SOFTWARE.
**/

#include "config.h"
#include "php_krb5.h"

#include "ext/standard/info.h"
#include "ext/standard/base64.h"

#include <sys/time.h>


/* Class definition */

PHP_METHOD(KRB5CCache, initPassword);
PHP_METHOD(KRB5CCache, initKeytab);
PHP_METHOD(KRB5CCache, getName);
PHP_METHOD(KRB5CCache, getEntries);
PHP_METHOD(KRB5CCache, open);
PHP_METHOD(KRB5CCache, save);
PHP_METHOD(KRB5CCache, isValid);
PHP_METHOD(KRB5CCache, setConfig);

static function_entry krb5_ccache_functions[] = {
		PHP_ME(KRB5CCache, initPassword, NULL, ZEND_ACC_PUBLIC)
		PHP_ME(KRB5CCache, initKeytab, NULL, ZEND_ACC_PUBLIC)
		PHP_ME(KRB5CCache, getName, NULL, ZEND_ACC_PUBLIC)
		PHP_ME(KRB5CCache, getEntries, NULL, ZEND_ACC_PUBLIC)
		PHP_ME(KRB5CCache, open, NULL, ZEND_ACC_PUBLIC)
		PHP_ME(KRB5CCache, save, NULL, ZEND_ACC_PUBLIC)
		PHP_ME(KRB5CCache, isValid, NULL, ZEND_ACC_PUBLIC)
		PHP_ME(KRB5CCache, setConfig, NULL, ZEND_ACC_PUBLIC)
		{NULL, NULL, NULL}
};


zend_module_entry krb5_module_entry = {
#if ZEND_MODULE_API_NO >= 20010901
    STANDARD_MODULE_HEADER,
#endif
    PHP_KRB5_EXT_NAME,
    NULL,
    PHP_MINIT(krb5),
    PHP_MSHUTDOWN(krb5),
    NULL,
    NULL,
    PHP_MINFO(krb5),
#if ZEND_MODULE_API_NO >= 20010901
    PHP_KRB5_EXT_VERSION,
#endif
    STANDARD_MODULE_PROPERTIES
};

#ifdef COMPILE_DL_KRB5
	ZEND_GET_MODULE(krb5)
#endif


krb5_error_code php_krb5_display_error(krb5_context ctx, krb5_error_code code, char* str TSRMLS_DC);

/*  Initialization functions */
zend_object_handlers krb5_ccache_handlers;
zend_object_value php_krb5_ticket_object_new( zend_class_entry *ce TSRMLS_DC);

PHP_MINIT_FUNCTION(krb5)
{
	zend_class_entry krb5_ccache;

	INIT_CLASS_ENTRY(krb5_ccache, "KRB5CCache", krb5_ccache_functions);
	krb5_ce_ccache = zend_register_internal_class(&krb5_ccache TSRMLS_CC);
	krb5_ce_ccache->create_object = php_krb5_ticket_object_new;
	memcpy(&krb5_ccache_handlers, zend_get_std_object_handlers(), sizeof(zend_object_handlers));

#ifdef HAVE_KADM5
	if(php_krb5_kadm5_register_classes(TSRMLS_C) != SUCCESS) {
		return FAILURE;
	}
#endif

	/* register constants */
	REGISTER_LONG_CONSTANT("GSS_C_DELEG_FLAG", GSS_C_DELEG_FLAG, CONST_CS | CONST_PERSISTENT );
	REGISTER_LONG_CONSTANT("GSS_C_MUTUAL_FLAG", GSS_C_MUTUAL_FLAG, CONST_CS | CONST_PERSISTENT );
	REGISTER_LONG_CONSTANT("GSS_C_REPLAY_FLAG", GSS_C_REPLAY_FLAG, CONST_CS | CONST_PERSISTENT );
	REGISTER_LONG_CONSTANT("GSS_C_SEQUENCE_FLAG", GSS_C_SEQUENCE_FLAG, CONST_CS | CONST_PERSISTENT );
	REGISTER_LONG_CONSTANT("GSS_C_CONF_FLAG", GSS_C_CONF_FLAG, CONST_CS | CONST_PERSISTENT );
	REGISTER_LONG_CONSTANT("GSS_C_INTEG_FLAG", GSS_C_INTEG_FLAG, CONST_CS | CONST_PERSISTENT );
	REGISTER_LONG_CONSTANT("GSS_C_ANON_FLAG", GSS_C_ANON_FLAG, CONST_CS | CONST_PERSISTENT );
	REGISTER_LONG_CONSTANT("GSS_C_PROT_READY_FLAG", GSS_C_PROT_READY_FLAG, CONST_CS | CONST_PERSISTENT );
	REGISTER_LONG_CONSTANT("GSS_C_TRANS_FLAG", GSS_C_TRANS_FLAG, CONST_CS | CONST_PERSISTENT );

	REGISTER_LONG_CONSTANT("GSS_C_BOTH", GSS_C_BOTH, CONST_CS | CONST_PERSISTENT );
	REGISTER_LONG_CONSTANT("GSS_C_INITIATE", GSS_C_INITIATE, CONST_CS | CONST_PERSISTENT );
	REGISTER_LONG_CONSTANT("GSS_C_ACCEPT", GSS_C_ACCEPT, CONST_CS | CONST_PERSISTENT );
	

	if(php_krb5_gssapi_register_classes(TSRMLS_C) != SUCCESS) {
		return FAILURE;
	}

	if(php_krb5_negotiate_auth_register_classes(TSRMLS_C) != SUCCESS) {
		return FAILURE;
	}

	return SUCCESS;
}

PHP_MSHUTDOWN_FUNCTION(krb5)
{
	if(php_krb5_gssapi_shutdown(TSRMLS_C) != SUCCESS) {
		return FAILURE;
	}

	return SUCCESS;
}


PHP_MINFO_FUNCTION(krb5)
{
	php_info_print_table_start();
	php_info_print_table_row(2, "Kerberos5 support", "enabled");
	php_info_print_table_row(2, "Version", PHP_KRB5_EXT_VERSION);
#ifdef HAVE_KRB5_MIT
	php_info_print_table_row(2, "Kerberos library", "MIT");
#endif


#ifdef HAVE_KRB5_HEIMDAL
	php_info_print_table_row(2, "Kerberos library", "Heimdal");
#endif


#ifdef HAVE_KADM5
	php_info_print_table_row(2, "KADM5 support", "yes");
#else

	php_info_print_table_row(2, "KADM5 support", "no");
#endif
	php_info_print_table_row(2, "GSSAPI/SPNEGO auth support", "yes");
	php_info_print_table_end();
}


/* compat functions */

#ifndef HAVE_KRB5_CC_NEW_UNIQUE

extern krb5_error_code KRB5_CALLCONV krb5_mcc_generate_new (krb5_context context, krb5_ccache *id);

/* {{{ */
krb5_error_code KRB5_CALLCONV krb5_cc_new_unique ( 
	krb5_context context, const char* type, const char* hint, krb5_ccache *id)
{
	krb5_error_code retval;	
	unsigned char random[16];
	unsigned char *rand64 = NULL;
	int rand64len = 0;

	/* generate a hopefully unique identifier of 128 bit length */

#ifdef HAVE_KRB5_RANDOM_MAKE_OCTETS
	krb5_data data;
	memset(&data, 0, sizeof(data));
	data.length = 16;
	data.data = (void*) &random;

	if((retval = krb5_c_random_make_octets(context,&data))) {
		return retval;
	}
	
#else
#ifndef HAVE_KRB5_RANDOM_CONFOUNDER
#error "No function to generate unique identifer found"
#endif
	if((retval = krb5_random_confounder(16, (krb5_pointer) &random))) {
		return retval;
	}
#endif

	
	rand64  = php_base64_encode((unsigned const char*) &random, 16, &rand64len);
	
	if(rand64 == NULL) {
		return KRB5_CRYPTO_INTERNAL;
	}

	char *ccname = malloc(strlen(type) + 2 + rand64len);

	*ccname = 0;
	strcat(ccname, type);
	strcat(ccname, ":");
	strncat(ccname, (const char*) rand64, rand64len);

	efree(rand64);

	if((retval = krb5_cc_resolve(context, ccname, id))) {
		free(ccname);
		return retval;
	}
	
	return retval;
}
/* }}} */


#endif

#ifndef HAVE_KRB5_GET_ERROR_MESSAGE

/* {{{ */
const char* KRB5_CALLCONV krb5_get_error_message (krb5_context context, krb5_error_code code)
{
	return error_message(code);
}
/* }}} */
#endif




/*  Constructors/Destructors */
/* {{{ */
static void php_krb5_ccache_object_dtor(void *obj, zend_object_handle handle TSRMLS_DC)
{
	krb5_ccache_object *ticket = (krb5_ccache_object*)obj;

	OBJECT_STD_DTOR(ticket->std);

	if(ticket) {
		krb5_cc_destroy(ticket->ctx, ticket->cc);
		krb5_free_context(ticket->ctx);

		if(ticket->keytab) {
			efree(ticket->keytab);
		}

		efree(ticket);
	}
} /* }}} */

/* {{{ */
zend_object_value php_krb5_ticket_object_new(zend_class_entry *ce TSRMLS_DC)
{
	zend_object_value retval;
	krb5_ccache_object *object;
	krb5_error_code ret = 0;

	object = emalloc(sizeof(krb5_ccache_object));
	memset(object, 0, sizeof(krb5_ccache_object));

	/* intialize context */
	if((ret = krb5_init_context(&object->ctx))) {
		zend_throw_exception(NULL, "Cannot initialize Kerberos5 context",0 TSRMLS_CC);
		return retval;
	}

	// initialize random ccache
	if((ret = krb5_cc_new_unique(object->ctx, "MEMORY", "", &object->cc))) {
		zend_throw_exception_ex(NULL, 0 TSRMLS_CC, "Cannot open credential cache (%s)", krb5_get_error_message(object->ctx,ret), ret);
	}


	INIT_STD_OBJECT(object->std, ce);

	zend_hash_copy(object->std.properties, &ce->default_properties,
					(copy_ctor_func_t) zval_add_ref, NULL,
					sizeof(zval*));

	retval.handle = zend_objects_store_put(object, php_krb5_ccache_object_dtor, NULL, NULL TSRMLS_CC);

	retval.handlers = &krb5_ccache_handlers;
	return retval;
} /* }}} */

/* Helper functions */
/* {{{ */
static int krb5_parse_init_creds_opts(zval *opts, krb5_get_init_creds_opt *cred_opts TSRMLS_DC)
{
	int retval = 0;
	zval **tmp = NULL;

	if (Z_TYPE_P(opts) != IS_ARRAY && Z_TYPE_P(opts) != IS_OBJECT) {
		return -1;
	}

	/* forwardable */
	if (zend_hash_find(HASH_OF(opts), "forwardable", sizeof("forwardable"), (void**) &tmp) == SUCCESS) {
		SEPARATE_ZVAL(tmp);
		convert_to_boolean_ex(tmp);

		krb5_get_init_creds_opt_set_forwardable(cred_opts, Z_LVAL_PP(tmp));

		zval_ptr_dtor(tmp);
	}

	/* proxiable */
	if (zend_hash_find(HASH_OF(opts), "proxiable", sizeof("proxiable"), (void**) &tmp) == SUCCESS) {
		SEPARATE_ZVAL(tmp);
		convert_to_boolean_ex(tmp);

		krb5_get_init_creds_opt_set_proxiable(cred_opts, Z_LVAL_PP(tmp));

		zval_ptr_dtor(tmp);
	}

	/* tkt_life */
	if (zend_hash_find(HASH_OF(opts), "tkt_life", sizeof("tkt_life"), (void**) &tmp) == SUCCESS) {
		SEPARATE_ZVAL(tmp);
		convert_to_long_ex(tmp);

		krb5_get_init_creds_opt_set_tkt_life(cred_opts, Z_LVAL_PP(tmp));

		zval_ptr_dtor(tmp);
	}

	/* renew_life */
	if (zend_hash_find(HASH_OF(opts), "renew_life", sizeof("renew_life"), (void**) &tmp) == SUCCESS) {
		SEPARATE_ZVAL(tmp);
		convert_to_long_ex(tmp);

		krb5_get_init_creds_opt_set_renew_life(cred_opts, Z_LVAL_PP(tmp));

		zval_ptr_dtor(tmp);
	}


	return retval;
} /* }}} */

/* {{{ */
krb5_error_code php_krb5_display_error(krb5_context ctx, krb5_error_code code, char* str TSRMLS_DC) {
	const char *errstr = krb5_get_error_message(ctx,code);

	zend_throw_exception_ex(NULL, 0 TSRMLS_CC, str, errstr);

	return code;
}
/* }}} */

/* {{{  Copies one ccache to another*/
static krb5_error_code php_krb5_copy_ccache(krb5_context ctx, const krb5_ccache src, krb5_ccache dest TSRMLS_DC)
{
	krb5_error_code retval = 0;
	krb5_principal princ;

	if((retval = krb5_cc_get_principal(ctx,src,&princ))) {
		return php_krb5_display_error(ctx, retval,  "Failed to retrive principal from source ccache (%s)" TSRMLS_CC);
	}

	if((retval = krb5_cc_initialize(ctx,dest,princ))) {
		krb5_free_principal(ctx, princ);
		return php_krb5_display_error(ctx, retval,  "Failed to initialize destination ccache (%s)" TSRMLS_CC);
	}

	krb5_free_principal(ctx, princ);


#ifdef HAVE_KRB5_HEIMDAL
	if((retval = krb5_cc_copy_cache(ctx, src, dest))) {
		return php_krb5_display_error(ctx, retval,  "Cannot copy given credential cache (%s)" TSRMLS_CC);
	}
#else
	krb5_cc_cursor cursor;
	if((retval = krb5_cc_start_seq_get(ctx,src,&cursor))) {
		return retval;
	}

	krb5_creds creds;
	while(krb5_cc_next_cred(ctx,src,&cursor,&creds) == 0) {
		if((retval = krb5_cc_store_cred(ctx, dest,&creds))) {
			krb5_cc_end_seq_get(ctx,src,&cursor);
			return retval;
		}
		krb5_free_cred_contents(ctx, &creds);
	}

	krb5_cc_end_seq_get(ctx,src,&cursor);
#endif

	return retval;
}
/* }}} */


/* KRB5CCache Methods */

/* {{{ proto string KRB5CCache::getName(  )
   Gets the name/identifier of this credential cache */
PHP_METHOD(KRB5CCache, getName)
{
	krb5_ccache_object *ccache = zend_object_store_get_object(getThis() TSRMLS_CC);

	const char *tmpname = krb5_cc_get_name(ccache->ctx, ccache->cc);
	const char *tmptype = krb5_cc_get_type(ccache->ctx, ccache->cc);
	char *name = NULL;

	name = malloc(strlen(tmpname) + strlen(tmptype) + 2);
	*name = 0;
	strcat(name, tmptype);
	strcat(name, ":");
	strcat(name, tmpname);

	RETVAL_STRING(name, 1);
	free(name);
} /* }}} */

/* {{{ proto void KRB5CCache::open( string $src )
   Copies the contents of the credential cache given by $dest to this credential cache */
PHP_METHOD(KRB5CCache, open)
{
	krb5_ccache_object *ccache = zend_object_store_get_object(getThis() TSRMLS_CC);
	char *sccname = NULL;
	size_t sccname_len = 0;
	krb5_error_code retval = 0;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s", &sccname, &sccname_len) == FAILURE) {
		return;
	}

	krb5_ccache src;

	if((retval = krb5_cc_resolve(ccache->ctx, sccname, &src))) {
		php_krb5_display_error(ccache->ctx, retval,  "Cannot open given credential cache (%s)" TSRMLS_CC);
		return;
	}

	if((retval = php_krb5_copy_ccache(ccache->ctx, src, ccache->cc TSRMLS_CC))) {
		php_krb5_display_error(ccache->ctx, retval,  "Failed to copy credential cache (%s)" TSRMLS_CC);
		return;
	}

	krb5_cc_close(ccache->ctx, src);
} /* }}} */

/* {{{ proto void KRB5CCache::save( string $dest )
   Copies the contents of this credential cache to the credential cache given by $dest */
PHP_METHOD(KRB5CCache, save)
{
	krb5_ccache_object *ccache = zend_object_store_get_object(getThis() TSRMLS_CC);
	char *sccname = NULL;
	size_t sccname_len = 0;
	krb5_error_code retval = 0;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s", &sccname, &sccname_len) == FAILURE) {
		return;
	}

	krb5_ccache dest = NULL;

	if((retval = krb5_cc_resolve(ccache->ctx, sccname, &dest))) {
		php_krb5_display_error(ccache->ctx, retval,  "Cannot open given credential cache (%s)" TSRMLS_CC);
		return;
	}

	if((retval = php_krb5_copy_ccache(ccache->ctx, ccache->cc, dest TSRMLS_CC))) {
		php_krb5_display_error(ccache->ctx, retval,  "Failed to copy credential cache (%s)" TSRMLS_CC);
		return;
	}

	krb5_cc_close(ccache->ctx, dest);
} /* }}} */

/* {{{ proto array KRB5CCache::initPassword( string $principal, string $pass [, array $options ])
   Gets a TGT for the given principal using the given credentials */
PHP_METHOD(KRB5CCache, initPassword)
{
	krb5_ccache_object *ccache = zend_object_store_get_object(getThis() TSRMLS_CC);
	krb5_creds creds;
	krb5_principal princ;
	krb5_error_code retval = 0;

	char *sprinc = NULL;
	int sprinc_len;
	char *spass = NULL;
	int spass_len;

	krb5_get_init_creds_opt cred_opts;
	memset(&cred_opts, 0, sizeof(cred_opts));
	krb5_get_init_creds_opt_init(&cred_opts);

	zval *opts = NULL;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ss|z", &sprinc, &sprinc_len, &spass, &spass_len, &opts) == FAILURE) {
		return;
	}

	if(opts) {
		if((retval = krb5_parse_init_creds_opts(opts, &cred_opts TSRMLS_CC))) {
			zend_throw_exception(NULL, "Cannot parse credential options", 0 TSRMLS_CC);
		}
	}

	if((retval = krb5_parse_name(ccache->ctx, sprinc, &princ))) {
		php_krb5_display_error(ccache->ctx, retval,  "Cannot parse Kerberos principal (%s)" TSRMLS_CC);
		return;
	}

	if((retval = krb5_get_init_creds_password(ccache->ctx, &creds, princ,
									spass, NULL, 0,	0, 0, &cred_opts))) {
		krb5_free_principal(ccache->ctx,princ);
		php_krb5_display_error(ccache->ctx, retval,  "Cannot get ticket (%s)" TSRMLS_CC);
		return;
	}

	if((retval = krb5_cc_initialize(ccache->ctx, ccache->cc, princ))) {
		krb5_free_principal(ccache->ctx,princ);
		php_krb5_display_error(ccache->ctx, retval,  "Failed to initialize credential cache (%s)" TSRMLS_CC);
		return;
	}

	krb5_free_principal(ccache->ctx,princ);

	if((retval = krb5_cc_store_cred(ccache->ctx, ccache->cc, &creds))) {
		php_krb5_display_error(ccache->ctx, retval,  "Failed to store ticket in credential cache (%s)" TSRMLS_CC);
		return;
	}

	krb5_free_cred_contents(ccache->ctx, &creds);
} /* }}} */

/* {{{ proto void KRB5CCache::initKeytab( string $principal, string $keytab_file [, array $options ])
   Gets a TGT for the given principal using the credentials in the given keytab */
PHP_METHOD(KRB5CCache, initKeytab)
{
	krb5_ccache_object *ccache = zend_object_store_get_object(getThis() TSRMLS_CC);
	krb5_keytab keytab;
	krb5_creds creds;
	krb5_principal princ;
	krb5_error_code retval = 0;

	char *sprinc = NULL;
	int sprinc_len;
	char *skeytab = NULL;
	int skeytab_len;

	krb5_get_init_creds_opt cred_opts;
	memset(&cred_opts, 0, sizeof(cred_opts));
	krb5_get_init_creds_opt_init(&cred_opts);

	zval *opts = NULL;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ss|z", &sprinc, &sprinc_len, &skeytab, &skeytab_len, &opts) == FAILURE) {
		return;
	}

#if PHP_MAJOR_VERSION < 6
	if((PG(safe_mode) && !php_checkuid(skeytab, NULL, CHECKUID_CHECK_FILE_AND_DIR)) ||
		php_check_open_basedir(skeytab TSRMLS_CC)) {
		return;
	}
#endif

	if(opts) {
		if((retval = krb5_parse_init_creds_opts(opts, &cred_opts TSRMLS_CC))) {
			zend_throw_exception(NULL, "Cannot parse credential options", 0 TSRMLS_CC);
		}
	}

	if((retval = krb5_parse_name(ccache->ctx, sprinc, &princ))) {
		php_krb5_display_error(ccache->ctx, retval,  "Cannot parse Kerberos principal (%s)" TSRMLS_CC);
		return;
	}

	if((retval = krb5_kt_resolve(ccache->ctx, skeytab, &keytab))) {
		krb5_free_principal(ccache->ctx,princ);
		php_krb5_display_error(ccache->ctx, retval,  "Cannot load keytab (%s)" TSRMLS_CC);
		return;
	}

	if((retval = krb5_get_init_creds_keytab(ccache->ctx, &creds, princ, keytab, 0,NULL, &cred_opts))) {
		krb5_free_principal(ccache->ctx,princ);
		php_krb5_display_error(ccache->ctx, retval,  "Cannot get ticket (%s)" TSRMLS_CC);
		return;
	}

	if((retval = krb5_cc_initialize(ccache->ctx, ccache->cc, princ))) {
		krb5_free_principal(ccache->ctx,princ);
		php_krb5_display_error(ccache->ctx, retval,  "Failed to initialize credential cache (%s)" TSRMLS_CC);
		return;
	}
	krb5_free_principal(ccache->ctx,princ);

	if((retval = krb5_cc_store_cred(ccache->ctx, ccache->cc, &creds))) {
		php_krb5_display_error(ccache->ctx, retval,  "Failed to store ticket in credential cache (%s)" TSRMLS_CC);
		return;
	}

	krb5_free_cred_contents(ccache->ctx, &creds);
	krb5_kt_close(ccache->ctx, keytab);

	ccache->keytab = estrdup(skeytab);
} /* }}} */

/* {{{ proto array KRB5CCache::getEntries( )
   Fetches all principal names for which tickets are available in this cache */
PHP_METHOD(KRB5CCache, getEntries)
{
	krb5_ccache_object *ccache = zend_object_store_get_object(getThis() TSRMLS_CC);
	krb5_error_code retval = 0;

	array_init(return_value);

	krb5_cc_cursor cursor;
	if((retval = krb5_cc_start_seq_get(ccache->ctx,ccache->cc,&cursor))) {
		php_krb5_display_error(ccache->ctx, retval,  "Failed to initialize ccache iterator (%s)" TSRMLS_CC);
		return;
	}

	krb5_creds creds;
	while(krb5_cc_next_cred(ccache->ctx,ccache->cc,&cursor,&creds) == 0) {
		if(creds.server) {
			char *princname = NULL;
			if((retval = krb5_unparse_name(ccache->ctx, creds.server, &princname))) {
				php_krb5_display_error(ccache->ctx, retval,  "Failed to unparse principal name (%s)" TSRMLS_CC);
			}
			add_next_index_string(return_value, princname, 1);

			free(princname);
			krb5_free_cred_contents(ccache->ctx, &creds);
		}
	}

	krb5_cc_end_seq_get(ccache->ctx,ccache->cc,&cursor);
} /* }}} */

/* {{{ proto bool KRB5CCache::isValid( [ int $timeRemain = 0 ] )
   Checks whether the credentials in the cache are still valid and remain valid for the given number of seconds */
PHP_METHOD(KRB5CCache, isValid)
{
	/* TODO: might be more useful if we were only checking the TGT */
	krb5_ccache_object *ccache = zend_object_store_get_object(getThis() TSRMLS_CC);
	krb5_error_code retval = 0;
	long time = 0;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "|l", &time) == FAILURE) {
		return;
	}

	krb5_cc_cursor cursor;
	if((retval = krb5_cc_start_seq_get(ccache->ctx,ccache->cc,&cursor))) {
		php_krb5_display_error(ccache->ctx, retval,  "Failed to initialize ccache iterator (%s)" TSRMLS_CC);
		return;
	}

	krb5_timestamp now = 0;

	if((retval = krb5_timeofday(ccache->ctx, &now))) {
		zend_throw_exception(NULL, "Failed to obtain time", 0 TSRMLS_CC);
	}

	krb5_creds creds;
	while(krb5_cc_next_cred(ccache->ctx,ccache->cc,&cursor,&creds) == 0) {
		krb5_timestamp validtill = creds.times.endtime - time;

		if(validtill <= now) {
			krb5_free_cred_contents(ccache->ctx, &creds);
			krb5_cc_end_seq_get(ccache->ctx,ccache->cc,&cursor);
			RETURN_FALSE;
		}
		krb5_free_cred_contents(ccache->ctx, &creds);
	}

	krb5_cc_end_seq_get(ccache->ctx,ccache->cc,&cursor);
	RETURN_TRUE;
} /* }}} */

/* {{{ proto void KRB5CCache::setConfig( string $config )
   Sets the kerberos configuration file to use (krb5.conf) */
PHP_METHOD(KRB5CCache, setConfig)
{
	krb5_ccache_object *ccache = zend_object_store_get_object(getThis() TSRMLS_CC);
	char *files[2] = { NULL , NULL };
	size_t file_len = 0;
	krb5_error_code retval = 0;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s", &files[0], &file_len) == FAILURE) {
			return;
	}

	if((retval = krb5_set_config_files(ccache->ctx, (const char**)&files))) {
		php_krb5_display_error(ccache->ctx, retval,  "Failed to set configuration file (%s)" TSRMLS_CC);
		return;
	}
} /* }}} */

